use std::mem;
use wgpu;

use super::BufferObject;

// Represents a textured vertex in WGSL
#[repr(C)]
#[derive(Copy, Clone, Debug, bytemuck::Pod, bytemuck::Zeroable)]
pub struct TexturedVertex {
	pub position: [f32; 3],
	pub texture_coords: [f32; 2],
}

impl BufferObject for TexturedVertex {
	fn layout() -> wgpu::VertexBufferLayout<'static> {
		wgpu::VertexBufferLayout {
			array_stride: Self::SIZE,
			step_mode: wgpu::VertexStepMode::Vertex,
			attributes: &[
				wgpu::VertexAttribute {
					offset: 0,
					shader_location: 0,
					format: wgpu::VertexFormat::Float32x3,
				},
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 3]>() as wgpu::BufferAddress,
					shader_location: 1,
					format: wgpu::VertexFormat::Float32x2,
				},
			],
		}
	}
}

// Represents a point in the starfield instance in WGSL
#[repr(C)]
#[derive(Copy, Clone, bytemuck::Pod, bytemuck::Zeroable)]
pub struct StarfieldInstance {
	/// Position in the starfield.
	///
	/// This is NOT world position, i.e, different from sprite positioning!
	/// The x and y coordinates here represent position relative to the center
	/// of a starfield tile in world units, which is converted to world position
	/// by the starfield vertex shader.
	pub position: [f32; 3],

	/// Star size, in world units. This does NOT scale with distance,
	/// unlike sprite size.
	pub size: f32,

	/// Parameters for this star's color variation,
	/// see the starfield fragment shader.
	pub tint: [f32; 2],
}

impl BufferObject for StarfieldInstance {
	fn layout() -> wgpu::VertexBufferLayout<'static> {
		wgpu::VertexBufferLayout {
			array_stride: Self::SIZE,
			step_mode: wgpu::VertexStepMode::Instance,
			attributes: &[
				// Position
				wgpu::VertexAttribute {
					offset: 0,
					shader_location: 2,
					format: wgpu::VertexFormat::Float32x3,
				},
				// Size
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 3]>() as wgpu::BufferAddress,
					shader_location: 3,
					format: wgpu::VertexFormat::Float32,
				},
				// Tint
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 4]>() as wgpu::BufferAddress,
					shader_location: 4,
					format: wgpu::VertexFormat::Float32x2,
				},
			],
		}
	}
}

// Represents a sprite instance in WGSL
#[repr(C)]
#[derive(Copy, Clone, bytemuck::Pod, bytemuck::Zeroable)]
pub struct SpriteInstance {
	/// Extra transformations this sprite
	/// (rotation, etc)
	pub transform: [[f32; 4]; 4],

	// What texture to use for this sprite
	pub texture_index: u32,
}

impl BufferObject for SpriteInstance {
	fn layout() -> wgpu::VertexBufferLayout<'static> {
		wgpu::VertexBufferLayout {
			array_stride: Self::SIZE,
			// Switch to a step mode of Vertex to Instance.
			// This means that our shaders will only change to use the next
			// instance when the shader starts processing a new instance
			step_mode: wgpu::VertexStepMode::Instance,
			attributes: &[
				// 4 arrays = 1 4x4 matrix
				wgpu::VertexAttribute {
					offset: 0,
					shader_location: 2,
					format: wgpu::VertexFormat::Float32x4,
				},
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 4]>() as wgpu::BufferAddress,
					shader_location: 3,
					format: wgpu::VertexFormat::Float32x4,
				},
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 8]>() as wgpu::BufferAddress,
					shader_location: 4,
					format: wgpu::VertexFormat::Float32x4,
				},
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 12]>() as wgpu::BufferAddress,
					shader_location: 5,
					format: wgpu::VertexFormat::Float32x4,
				},
				// Texture
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 16]>() as wgpu::BufferAddress,
					shader_location: 6,
					format: wgpu::VertexFormat::Uint32,
				},
			],
		}
	}
}
